from __future__ import annotations

import dataclasses
import typing
from http import HTTPMethod, HTTPStatus
from reprlib import recursive_repr

if typing.TYPE_CHECKING:
    from saronia.auth import AuthError

type BaseError = AuthError | STATUS_ERROR | NetworkError | UnknownError

STATUS_ERROR: typing.Final = type("StatusError", (Exception,), {"__module__": __name__})


def status(code: int, description: str = "", /) -> tuple[HTTPStatus, str]:
    http_status = HTTPStatus(code)
    return (http_status, description or http_status.description)


def status_error(code: int, description: str = "", /) -> BaseStatusError:
    return StatusError[status(code, description)]  # type: ignore


@dataclasses.dataclass(frozen=True, slots=True)
class UnknownError:
    message: bytes
    orig_error: BaseException | None = None


class NetworkError(Exception):
    """Raised when a network-level error occurs (connection failed, timeout, etc)."""

    __match_args__ = ("message", "original_error")

    def __init__(self, message: str, original_error: BaseException | None = None) -> None:
        self.message = message
        self.original_error = original_error
        super().__init__(message)

    def __str__(self) -> str:
        if self.original_error:
            return f"{self.message}: {self.original_error}"
        return self.message


class BaseStatusError:
    _errors: dict[HTTPStatus, STATUS_ERROR]

    @classmethod
    def get_status_error(cls, status: HTTPStatus, /) -> STATUS_ERROR | None:
        return cls._errors.get(status)


class StatusError(STATUS_ERROR):
    if typing.TYPE_CHECKING:
        STATUSES: typing.ClassVar[tuple[HTTPStatus, ...]]

    def __class_getitem__(
        cls,
        statuses: HTTPStatus | tuple[HTTPStatus, ...] | tuple[tuple[HTTPStatus, str], ...],
        /,
    ) -> typing.Any:
        statuses = (statuses,) if not isinstance(statuses, tuple) else statuses

        if statuses and (isinstance(statuses[0], tuple)) or (len(statuses) == 2 and isinstance(statuses[0], HTTPStatus) and isinstance(statuses[1], str)):
            statuses_with_descriptions = ((statuses[0], statuses[1]),) if len(statuses) == 2 and not isinstance(statuses[0], tuple) else statuses
            return type(
                cls.__name__,
                (BaseStatusError,),
                {
                    "_errors": {
                        http_status: StatusError(description)
                        for http_status, description in typing.cast("tuple[tuple[HTTPStatus, str]]", statuses_with_descriptions)
                    },
                },
            )

        return type(cls.__name__ + "Mixin", (), {"STATUSES": statuses, "__module__": cls.__module__})


class APIError[Error = typing.Never](Exception):
    """Represents an API error response.

    Attributes:
        error: The parsed error object
        method: The HTTP method that was used
        status: The HTTP status code returned
        path: The request path (optional)
        request_id: Request ID for tracing (optional)

    """

    __match_args__ = ("error", "method", "status", "path", "request_id")

    error: Error | BaseError

    def __init__(
        self,
        error: Error | BaseError,
        method: HTTPMethod,
        status: HTTPStatus,
        *,
        path: str | None = None,
        request_id: str | None = None,
    ) -> None:
        self.error = error
        self.method = method
        self.status = status
        self.path = path
        self.request_id = request_id
        super().__init__()

    def __str__(self) -> str:
        return self.__repr__()

    @recursive_repr()
    def __repr__(self) -> str:
        parts = [f"{type(self).__name__} [{self.status.value}] ({self.status.phrase})"]

        if self.path:
            parts.append(f"path={self.path}")

        if self.request_id:
            parts.append(f"request_id={self.request_id}")

        parts.append(f"error={self.error!r}")
        return f"<{' '.join(parts)}>"


__all__ = ("APIError", "NetworkError", "StatusError", "UnknownError", "status", "status_error")



# pyright: reportAttributeAccessIssue=false, reportMissingImports=false

import datetime
import typing
from http import HTTPMethod, HTTPStatus
from io import IOBase

from kungfu import Ok, Option
from msgspex import decoder, encoder

from saronia.client.base import DEFAULT_TIMEOUT, DEFAULT_USER_AGENT, BaseClient, MultipartFile

if typing.TYPE_CHECKING:
    import rnet


class RnetClient(BaseClient):
    def __init__(
        self,
        client: "rnet.Client",
        base_url: str = "",
        *,
        user_agent: str | None = None,
        request_timeout: datetime.timedelta | float = DEFAULT_TIMEOUT,
        default_headers: bool = False,
    ) -> None:
        super().__init__(
            user_agent=user_agent or DEFAULT_USER_AGENT.format(http_client="rnet3"),
            base_url=base_url.rstrip("/"),
        )

        self.client = client
        self.default_headers = default_headers
        self.request_timeout = (
            request_timeout
            if request_timeout is None
            else datetime.timedelta(seconds=request_timeout)
            if isinstance(request_timeout, int | float)
            else request_timeout
        )

    async def request(
        self,
        path: str,
        method: HTTPMethod,
        *,
        as_result: bool = False,
        errors: tuple[typing.Any, ...],
        response_type: Option[typing.Any],
        json: Option[str | bytes],
        headers: Option[typing.Mapping[str, typing.Any]],
        urlencoded_params: Option[typing.Mapping[str, typing.Any]],
        query_params: Option[typing.Mapping[str, typing.Any]],
        body: Option[typing.Any],
        files: Option[typing.Mapping[str, MultipartFile]],
        auth: typing.Any = None,
    ) -> typing.Any:
        import rnet
        import rnet.exceptions

        url = f"{self.base_url}{path}" if self.base_url else path
        status = request_id = None

        try:
            kwargs: dict[str, typing.Any] = {
                "headers": self.headers.copy(),
                "query": self.query_parameters.copy(),
                "cookies": self.cookies.copy(),
            }

            self._apply_auth(auth, kwargs["headers"], kwargs["query"], kwargs["cookies"])

            if headers:
                kwargs["headers"] |= {k.title(): v if isinstance(v, str) else encoder.encode(v).strip('"') for k, v in headers.unwrap().items()}

            if query_params:
                kwargs["query"] |= {k: v if isinstance(v, str | int | float | bool) else encoder.encode(v).strip('"') for k, v in query_params.unwrap().items()}

            if json:
                json_data = json.unwrap()
                kwargs["headers"]["Content-Type"] = "application/json"
                kwargs["body"] = json_data if isinstance(json_data, bytes) else json_data.encode()
            elif body:
                kwargs["body"] = body.unwrap()
            elif urlencoded_params and not files:
                kwargs["form"] = {
                    k: v if isinstance(v, str | int | bool | float) else encoder.encode(v).strip('"') for k, v in urlencoded_params.unwrap().items()
                }
            elif files:
                parts: list[rnet.Part] = []

                for k, v in urlencoded_params.unwrap().items():
                    parts.append(rnet.Part(k, v if isinstance(v, str) else encoder.encode(v).strip('"')))

                for field_name, file_info in files.unwrap().items():
                    content = file_info.content.read() if isinstance(file_info.content, typing.IO | IOBase) else file_info.content
                    parts.append(
                        rnet.Part(
                            name=field_name,
                            value=content,
                            filename=file_info.name,
                            mime=file_info.mime,
                        ),
                    )

                kwargs["multipart"] = rnet.Multipart(*parts)

            resp = await self.client.request(
                getattr(rnet.Method, method.name),
                url,
                default_headers=self.default_headers,
                timeout=self.request_timeout,
                **kwargs,
            )

            if 200 <= resp.status.as_int() < 300:
                response = decoder.decode(await resp.bytes(), type=response_type.unwrap_or(typing.Any))
                return Ok(response) if as_result else response

            status = HTTPStatus(resp.status.as_int())
            request_id = None if not (req_id := resp.headers.get("x-request-id") or resp.headers.get("request-id")) else req_id.decode()

            self._raise_error(
                path,
                method,
                status=status,
                payload=await resp.bytes(),
                errors=errors,
                request_id=request_id,
            )
        except SystemExit, KeyboardInterrupt:
            raise
        except BaseException as exception:
            return self._handle_error(
                status,
                method,
                path,
                request_id,
                exception,
                rnet.exceptions.ConnectionError,
                rnet.exceptions.ConnectionResetError,
                rnet.exceptions.ProxyConnectionError,
                rnet.exceptions.RequestError,
                rnet.exceptions.TimeoutError,
                rnet.exceptions.RustPanic,
                rnet.exceptions.TlsError,
                rnet.exceptions.BodyError,
                rnet.exceptions.RedirectError,
                as_result=as_result,
            )


__all__ = ("RnetClient",)



# pyright: reportAttributeAccessIssue=false, reportMissingImports=false

import typing
from http import HTTPMethod, HTTPStatus

from kungfu import Ok, Option
from kungfu.library.monad.option import NOTHING
from msgspex import decoder, encoder

from saronia.client.base import DEFAULT_TIMEOUT, DEFAULT_USER_AGENT, BaseClient, MultipartFile

if typing.TYPE_CHECKING:
    import aiohttp


class AiohttpClient(BaseClient):
    def __init__(
        self,
        session: "aiohttp.ClientSession",
        *,
        user_agent: str | None = None,
        request_timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        from aiohttp import __version__

        super().__init__(
            user_agent=user_agent or DEFAULT_USER_AGENT.format(http_client=f"aiohttp/{__version__}"),
        )

        self.session = session
        self.request_timeout = request_timeout

    async def request(
        self,
        path: str,
        method: HTTPMethod,
        *,
        as_result: bool,
        errors: tuple[typing.Any, ...],
        response_type: Option[typing.Any],
        json: Option[str | bytes],
        headers: Option[typing.Mapping[str, typing.Any]],
        urlencoded_params: Option[typing.Mapping[str, typing.Any]],
        query_params: Option[typing.Mapping[str, typing.Any]],
        body: Option[typing.Any],
        files: Option[typing.Mapping[str, MultipartFile]],
        auth: typing.Any = None,
    ) -> typing.Any:
        import aiohttp
        import aiohttp.client_exceptions
        import aiohttp.http_exceptions

        status = request_id = None

        try:
            kwargs: dict[str, typing.Any] = {
                "headers": self.headers.copy(),
                "params": self.query_parameters.copy(),
                "cookies": self.cookies.copy(),
            }

            self._apply_auth(auth, kwargs["headers"], kwargs["params"], kwargs["cookies"])

            if headers:
                kwargs["headers"] = {k.title(): v if isinstance(v, str) else encoder.encode(v).strip('"') for k, v in headers.unwrap().items()}

            if query_params:
                kwargs["params"] = {k: v if isinstance(v, str | int | float | bool) else encoder.encode(v).strip('"') for k, v in query_params.unwrap().items()}

            if json:
                json_data = json.unwrap()
                kwargs["headers"]["Content-Type"] = "application/json"
                kwargs["data"] = json_data if isinstance(json_data, bytes) else json_data.encode()
            elif body is not NOTHING:
                kwargs["data"] = body.unwrap()
            elif urlencoded_params and not files:
                kwargs["data"] = dict(urlencoded_params.unwrap())
            elif files:
                form_data = aiohttp.FormData(quote_fields=False)

                for k, v in urlencoded_params.unwrap().items():
                    form_data.add_field(k, v if isinstance(v, str) else encoder.encode(v).strip('"'))

                for field_name, file_info in files.unwrap().items():
                    form_data.add_field(
                        field_name,
                        file_info.content,
                        filename=file_info.name,
                        content_type=file_info.mime or "application/octet-stream",
                    )

                kwargs["data"] = form_data

            async with self.session.request(method.value, path, **kwargs) as resp:
                if 200 <= resp.status < 300:
                    return Ok(decoder.decode(await resp.read(), type=response_type.unwrap_or(typing.Any)))

                status = HTTPStatus(resp.status)
                request_id = resp.headers.get("X-Request-ID") or resp.headers.get("Request-ID")

                self._raise_error(
                    path,
                    method,
                    status=status,
                    payload=await resp.read(),
                    errors=errors,
                    request_id=request_id,
                )
        except SystemExit, KeyboardInterrupt:
            raise
        except BaseException as exception:
            return self._handle_error(
                status,
                method,
                path,
                request_id,
                exception,
                TimeoutError,
                aiohttp.client_exceptions.ClientError,
                aiohttp.http_exceptions.HttpProcessingError,
                as_result=as_result,
            )


__all__ = ("AiohttpClient",)



import abc
import dataclasses
import platform
import sys
import typing
from contextlib import suppress
from http import HTTPMethod, HTTPStatus

from kungfu import Error, Option
from msgspex import decoder

from saronia.__meta__ import __version__
from saronia.auth import AuthError
from saronia.client.abc import ABCClient, MultipartFile
from saronia.error import STATUS_ERROR, APIError, BaseStatusError, NetworkError, UnknownError

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

    from saronia.auth import Auth, AuthComposite

_SENTINEL: typing.Final = object()
DEFAULT_TIMEOUT: typing.Final = 30.0
DEFAULT_USER_AGENT: typing.Final = "CPython/{py_major}.{py_minor}; ({system}; {platform}) {saronia} {http_client}".format(
    py_major=sys.version_info.major,
    py_minor=sys.version_info.minor,
    system=platform.system() or "Unknown",
    platform=platform.platform(terse=True) or "Unknown",
    saronia=f"saronia/{__version__}",
    http_client="{http_client}",
)


class BaseClient(ABCClient, abc.ABC):
    base_url: str
    headers: dict[str, str]
    query_parameters: dict[str, str]
    cookies: dict[str, str]
    auth_model: DataclassInstance | None

    def __init__(self, user_agent: str, base_url: str = "") -> None:
        self.base_url = base_url
        self.headers = {"User-Agent": user_agent}
        self.query_parameters = {}
        self.cookies = {}
        self.auth_model = None

    def auth(self, auth_model: DataclassInstance) -> None:
        self.auth_model = auth_model

    @abc.abstractmethod
    async def request(
        self,
        path: str,
        method: HTTPMethod,
        *,
        as_result: bool,
        errors: tuple[typing.Any, ...],
        response_type: Option[typing.Any],
        json: Option[str | bytes],
        headers: Option[typing.Mapping[str, typing.Any]],
        urlencoded_params: Option[typing.Mapping[str, typing.Any]],
        query_params: Option[typing.Mapping[str, typing.Any]],
        body: Option[typing.Any],
        files: Option[typing.Mapping[str, MultipartFile]],
        auth: typing.Any = None,
    ) -> typing.Any:
        pass

    def _handle_error(
        self,
        status: HTTPStatus | None,
        method: HTTPMethod,
        path: str,
        request_id: str | None,
        exception: BaseException,
        /,
        *http_errors: type[BaseException],
        as_result: bool = False,
    ) -> Error[APIError[typing.Any]]:
        api_error_kwargs: dict[str, typing.Any] = dict(
            method=method,
            path=path,
            request_id=request_id,
        )

        if isinstance(exception, http_errors):
            error = APIError(
                error=NetworkError("Network issue", exception),
                status=status or HTTPStatus.BAD_REQUEST,
                **api_error_kwargs,
            )

        elif isinstance(exception, (AuthError, STATUS_ERROR)):
            error = APIError(
                error=exception,
                status=status or HTTPStatus.NETWORK_AUTHENTICATION_REQUIRED,
                **api_error_kwargs,
            )

        elif isinstance(exception, APIError):
            error = exception  # type: ignore

        else:
            error = APIError(
                error=UnknownError(b"", exception),
                status=status or HTTPStatus.BAD_REQUEST,
                **api_error_kwargs,
            )

        if as_result:
            return Error(error)

        raise error

    def _raise_error(
        self,
        path: str,
        method: HTTPMethod,
        status: HTTPStatus,
        payload: bytes,
        errors: tuple[typing.Any, ...],
        request_id: str | None = None,
    ) -> typing.NoReturn:
        for error_type in errors:
            if isinstance(error_type, type) and issubclass(error_type, BaseStatusError):
                status_error = error_type.get_status_error(status)

                if status_error is not None:
                    raise status_error

                continue

            if payload:
                statuses = getattr(error_type, "STATUSES", ())

                if statuses and status not in statuses:
                    continue

                error_obj = _SENTINEL

                with suppress(Exception):
                    error_obj = decoder.decode(payload, type=error_type)

                if error_obj is not _SENTINEL:
                    raise APIError(error_obj, method, status, path=path, request_id=request_id)

        raise APIError(UnknownError(payload), method, status, path=path, request_id=request_id)

    def _apply_auth(
        self,
        auth: Auth | AuthComposite | None,
        headers: dict[str, str],
        query_params: dict[str, str],
        cookies: dict[str, str],
    ) -> None:
        if auth is None:
            return

        from saronia.auth import AuthComposite, AuthError
        from saronia.security import CookieAPIKey, HeaderAPIKey, HTTPAuthorization, QueryAPIKey

        if isinstance(auth, type):
            if self.auth_model is None:
                raise AuthError(f"Auth `{auth.__name__}` required but no credentials provided")

            for field in dataclasses.fields(self.auth_model):
                field_value = getattr(self.auth_model, field.name)
                if field_value is not None and type(field_value) is auth:
                    auth = field_value
                    break
            else:
                raise AuthError(f"Auth `{auth.__name__}` is not defined in `{self.auth_model.__class__.__name__}`")

        if isinstance(auth, AuthComposite):
            match auth.op:
                case "AND":
                    self._apply_auth(auth.left, headers, query_params, cookies)
                    self._apply_auth(auth.right, headers, query_params, cookies)
                case "OR":
                    try:
                        self._apply_auth(auth.left, headers, query_params, cookies)
                        return
                    except AuthError:
                        self._apply_auth(auth.right, headers, query_params, cookies)
                case "NOT":
                    raise AuthError("NOT operator requires combination with other auth")
                case _:
                    typing.assert_never(auth.op)

            return

        match auth:
            case CookieAPIKey():
                cookies.update(auth.mapping)
            case HeaderAPIKey():
                headers.update(auth.mapping)
            case QueryAPIKey():
                query_params.update(auth.mapping)
            case HTTPAuthorization():
                headers.update(auth.header)
            case _:
                raise AuthError(f"Unknown auth type: `{type(auth)}`")


__all__ = ("BaseClient",)
